#include <math.h>
#include <X11/keysym.h>
#include <Xm/ManagerP.h>
#include "KnobGP.h"

/* defines for mathmatic functions used in drawing */
#define  RADIANS(x)  (M_PI * 2.0 * (x) / 360.0)
#define  DEGREES(x)  ((x) / (M_PI * 2.0) * 360.0)
#define  MIN_ANGLE   225.0
#define  MAX_ANGLE   345.0
#define  NUM_SEGS    24
#define  MIN(a,b)    (((a) < (b)) ? (a) :  (b))

#define MIN_KNOB_DIAMETER                3
#define DEFAULT_DIAMETER                100
#define DEFAULT_KNOB_MARGIN                5
#define DEFAULT_KNOB_MARGIN_PERCENTAGE        50

/********    Static Function Declarations    ********/

/* rectobj methods */
static void ClassPartInitialize( WidgetClass widget_class) ;
static void Initialize( Widget request, Widget new_w, ArgList args,
                        Cardinal *num_args) ;
static void Destroy( Widget w) ;
static void Resize( Widget w) ;
static void Redisplay( Widget w, XEvent *event, Region region) ;
static Boolean SetValues( Widget old_w, Widget request, Widget new_w,
                          ArgList args, Cardinal *num_args) ;
static void SetValuesAlmost( Widget old_w, Widget new_w,
                            XtWidgetGeometry *request, XtWidgetGeometry *reply);
static XtGeometryResult QueryGeometry( Widget w, XtWidgetGeometry *request, 
                                            XtWidgetGeometry *reply) ;

static void GetDiameters( Widget w, unsigned int *diameter,
                         unsigned int *inner_diameter) ;

/* gadget methods */
static void HighlightKnob( Widget w) ;
static void UnhighlightKnob( Widget w) ;
static void InputDispatch( Widget wid, XEvent *event, Mask event_mask) ;
static void ArmAndActivate( Widget wid, XEvent *event, String *params,
                            Cardinal *num_params) ;
static Boolean VisualChange( Widget wid, Widget cmw, Widget nmw) ;


/* knob methods */
static void CreateSegments( Widget w) ;
static void DrawIndicator( Widget w) ;
static void TurnKnob( XtPointer closure, XtIntervalId *id) ;
static void DrawKnob( Widget w) ;

/* resource callprocs */
static void DefaultIndicatorColor( Widget, int, XrmValue *);
static void DefaultKnobColor( Widget, int, XrmValue *);

/* action procs */
static void HandleKey( Widget w, XEvent *event, char **params,
                      Cardinal *num_params) ;
static void ToggleLeft( Widget w, XEvent *event, char **params,
                        Cardinal *num_params) ;
static void ToggleRight( Widget w, XEvent *event, char **params,
                         Cardinal *num_params) ;
static void ToggleKnob( Widget w, XEvent *event, char **params,
                      Cardinal *num_params) ;
static void StartKnobTurn( Widget w, XEvent *event, char **params,
                         Cardinal *num_params) ;
static void ReleaseKnob( Widget w, XEvent *event, char **params,
                         Cardinal *num_params) ;

/* Resource list initialization - sets up resource names and values
 * for each resource of this widget.
 */
static XtResource resources[] = {
        {
          XsmNvalueChangedCallback, XmCCallback, XmRCallback,
          sizeof(XtCallbackList),
          XtOffset (XsmKnobGadget, knobg.value_changed_callback),
          XmRImmediate, NULL
        },

        {
          XsmNvalue, XsmCValue, XmRInt, sizeof(int),
          XtOffset (XsmKnobGadget, knobg.value),
          XmRImmediate, (XtPointer) 0
        },
          
        {
          XsmNmaxValue, XsmCMaxValue, XmRInt, sizeof(int),
          XtOffset (XsmKnobGadget, knobg.max_val),
          XmRImmediate, (XtPointer) 100
        },
          
        {
          XsmNminValue, XsmCMinValue, XmRInt, sizeof(int),
          XtOffset (XsmKnobGadget, knobg.min_val),
          XmRImmediate, (XtPointer) 0
        },
          
        {
          XsmNturnDelay, XsmCTurnDelay, XmRInt, sizeof(int),
          XtOffset(XsmKnobGadget, knobg.turn_delay),
          XmRImmediate, (XtPointer) 50
        },

        {
          XsmNknobColor, XmCBackground, XmRPixel, sizeof (Pixel),
          XtOffset (XsmKnobGadget, knobg.knob_color),
          XmRCallProc, (XtPointer) DefaultKnobColor
        },
          
        {
          XsmNindicatorColor, XsmCIndicatorColor, XmRPixel, sizeof (Pixel),
          XtOffset (XsmKnobGadget, knobg.indicator_color),
          XmRCallProc, (XtPointer) DefaultIndicatorColor
        },

        {
          XsmNmarginWidth, XmCMarginWidth, XmRHorizontalDimension,
          sizeof (Dimension),
          XtOffset (XsmKnobGadget, knobg.margin_width),
          XmRImmediate, (XtPointer) DEFAULT_KNOB_MARGIN
        },
          
        {
          XsmNmarginHeight, XmCMarginHeight, XmRVerticalDimension,
          sizeof (Dimension),
          XtOffset (XsmKnobGadget, knobg.margin_height),
          XmRImmediate, (XtPointer) DEFAULT_KNOB_MARGIN
        },
          
        {
          XsmNknobMargin, XsmCKnobMargin, XmRDimension, sizeof (Dimension),
          XtOffset (XsmKnobGadget, knobg.knob_margin),
          XmRImmediate, (XtPointer) DEFAULT_KNOB_MARGIN_PERCENTAGE
        },
};
          
/* Synthetic resource list initialization - sets up resource procedures
 * to get called when setting or getting the various resources.  These
 * synthetic resources are used in resolution independance.
 */
static XmSyntheticResource syn_resources[] = {
        {
          XsmNmarginWidth,
          sizeof (Dimension),
          XtOffset( XsmKnobGadget, knobg.margin_width),
          _XmFromHorizontalPixels,
          _XmToHorizontalPixels
        },

        {
          XsmNmarginHeight,
          sizeof (Dimension),
          XtOffset( XsmKnobGadget, knobg.margin_height),
          _XmFromVerticalPixels,
          _XmToVerticalPixels
        }
};


/* Widget class record initialization - Here we initialize each field of the
 * rectobj, gadget, and knob class parts.
 */
externaldef (xsmknobgadgetclassrec)
            XsmKnobGadgetClassRec xsmKnobGadgetClassRec = {
   {
						/* RectObj Class Rec   */
      (WidgetClass) &xmGadgetClassRec,          /* superclass          */
      "XsmKnobGadget",                          /* class_name          */
      sizeof(XsmKnobGadgetRec),                 /* widget_size         */
      NULL,                                     /* class_initialize    */
      ClassPartInitialize,                      /* class_part_initiali */
      FALSE,                                    /* class_inited        */
      Initialize,                               /* initialize          */
      NULL,                                     /* initialize_hook     */
      NULL,                                     /* realize             */
      NULL,                                     /* actions             */
      0,                                        /* num_actions         */
      resources,                                /* resources           */
      XtNumber(resources),                      /* num_resources       */
      NULLQUARK,                                /* xrm_class           */
      TRUE,                                     /* compress_motion     */
      XtExposeCompressMaximal,                  /* compress_exposure   */
      TRUE,                                     /* compress_enterleave */
      FALSE,                                    /* visible_interest    */
      Destroy,                                  /* destroy             */
      Resize,                                   /* resize              */
      Redisplay,                                /* expose              */
      SetValues,                                /* set_values          */
      NULL,                                     /* set_values_hook     */
      SetValuesAlmost,                          /* set_values_almost   */
      NULL,                                     /* get_values_hook     */
      NULL,                                     /* accept_focus        */
      XtVersion,                                /* version             */
      NULL,                                     /* callback_private    */
      NULL,                                     /* tm_table            */
      QueryGeometry,                            /* query_geometry      */
      NULL,                                     /* display accel       */
      NULL,                                     /* extension           */
   },

   {                                            /* Gadget Class Rec    */
      HighlightKnob,                            /* border_highlight    */
      UnhighlightKnob,                          /* border_unhighlight  */
      ArmAndActivate,                           /* arm_and_activate    */
      InputDispatch,                            /* input dispatch      */
      VisualChange,                             /* visual_change       */
      syn_resources,                            /* syn resources       */
      XtNumber(syn_resources),                  /* num syn_resources   */
      NULL,				        /* class cache part    */
      NULL,                                     /* extension           */
   },

   {                                            /* Knob Gadget Class Rec */
      GetDiameters,                             /* get_diameters         */
      CreateSegments,                           /* create_segments       */
      DrawIndicator,                            /* draw_indicator        */
      TurnKnob,                                 /* turn                  */
      DrawKnob,                                 /* draw                  */
      NULL,                                     /* extension             */
   }
};

externaldef(xsmknobwidgetclass) WidgetClass xsmKnobGadgetClass =
                                         (WidgetClass) &xsmKnobGadgetClassRec;

/****************************************************************
 *
 * Class Functions
 *
 ****************************************************************/


/************************************************************************
 *
 *  ClassPartInitialize() -
 *      The initialization routine for class parts.  It is called for the
 *        initialization of this class and each time a new subclass is created
 *        for this widget class.  It is called only once per subclass.
 *
 ************************************************************************/
static void ClassPartInitialize(WidgetClass wc)
{
    XsmKnobGadgetClass kc = (XsmKnobGadgetClass) wc;
    XsmKnobGadgetClass sc = (XsmKnobGadgetClass) wc->core_class.superclass;

   /* assign procedures to the classes pointers that inherit these procedures */

    if (kc->knobg_class.get_diameters == XsmInheritGetDiameters)
       kc->knobg_class.get_diameters = sc->knobg_class.get_diameters;

    if (kc->knobg_class.create_segments == XsmInheritCreateSegments)
       kc->knobg_class.create_segments = sc->knobg_class.create_segments;

    if (kc->knobg_class.draw_indicator == XsmInheritDrawIndicator)
       kc->knobg_class.draw_indicator = sc->knobg_class.draw_indicator;

    if (kc->knobg_class.turn == XsmInheritTurn)
       kc->knobg_class.turn = sc->knobg_class.turn;

    if (kc->knobg_class.draw == XsmInheritDraw)
       kc->knobg_class.draw = sc->knobg_class.draw;

}



/************************************************************************
 *
 *  Initialize() -
 *      The main widget instance initialization routine.  Verify resource
 *  values and initialize widget instance fields.
 *
 ************************************************************************/

/* ARGSUSED */
static void Initialize( Widget request, Widget new_w, ArgList args,
                        Cardinal *num_args )
{
    XsmKnobGadget req_kg = (XsmKnobGadget) request;
    XsmKnobGadget new_kg = (XsmKnobGadget) new_w;
    XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(new_w);
    XmManagerWidget mw = (XmManagerWidget) XtParent(new_w);
    Dimension min_width, min_height;
    XGCValues values;
    unsigned long valuemask;

   /* Verify resource values */

   /* verify that the maximum value is not less than or equal to zero */
    if (new_kg->knobg.max_val <= 0) {
      XtWarning("MaxValue is must be greater than 0, Defaulting MaxValue to 1");
      new_kg->knobg.max_val = 1;
    }

   /*
    * verify that the maximum value is not less than the minimum value
    */
    if (new_kg->knobg.max_val < new_kg->knobg.min_val) {
       XtWarning("MaxValue is less than the MinValue, Defaulting to MinValue to 0");
       new_kg->knobg.min_val = 0;
    }

   /* verify that the value is not greater than the maximum value */
    if (new_kg->knobg.value > new_kg->knobg.max_val) {
       XtWarning("Value is greater than MaxValue, Defaulting to MaxValue");
       new_kg->knobg.value = new_kg->knobg.max_val;
    }

   /* verify that the value is not less than the minimum value */
    if (new_kg->knobg.value < new_kg->knobg.min_val) {
       XtWarning("Value is less than MinValue, Defaulting to MinValue");
       new_kg->knobg.value = new_kg->knobg.min_val;
    }

   /* verify that turn rate is not less than 1 */
    if (new_kg->knobg.turn_delay < 1) {
       XtWarning("Turn Delay is less than 1, Defaulting to 1");
       new_kg->knobg.turn_delay = 1;
    }

   /*
    * Knob margin is a percentage of the total knob,
    * thus it must be between 0 and 100.  Since this resource is
    * typed as a dimension (unsigned short), a negative value
    * becomes a large positive, thus this also checks for < 0/
    */

    if (new_kg->knobg.knob_margin > 100)
    {
       XtWarning("Knob Margin must be between 0 and 100");
       new_kg->knobg.knob_margin = DEFAULT_KNOB_MARGIN_PERCENTAGE;
    }
       

   /*
    * initialize the minimum required dimensions for the knob
    */

    min_width = MIN_KNOB_DIAMETER +
                (2 * (new_kg->knobg.margin_width +
                      new_kg->gadget.highlight_thickness));
    min_height = MIN_KNOB_DIAMETER +
                 (2 * (new_kg->knobg.margin_width +
                       new_kg->gadget.highlight_thickness));
                
   /*
    * if an initial width and height weren't set, initialize
    * a default dimensions, otherwise verify that the dimensions
    * that were set are at least the minimum dimension
    */

    if (req_kg->rectangle.width == 0)
       new_kg->rectangle.width = DEFAULT_DIAMETER +
                            (2 * (new_kg->knobg.margin_width +
                                  new_kg->gadget.highlight_thickness));
    else
       if (new_kg->rectangle.width < min_width)
          new_kg->rectangle.width = min_width;

    if (req_kg->rectangle.height == 0)
       new_kg->rectangle.height = DEFAULT_DIAMETER +
                            (2 * (new_kg->knobg.margin_height +
                                  new_kg->gadget.highlight_thickness));
    else
       if (new_kg->rectangle.height < min_width)
          new_kg->rectangle.height = min_height;



   /* save original calculated width and height for use in query geometry */
       new_kg->knobg.orig_width = new_kg->rectangle.width;
       new_kg->knobg.orig_height = new_kg->rectangle.height;

   /* Initialize internal variables */

    new_kg->knobg.num_segments = NUM_SEGS;
    new_kg->knobg.segments = (XSegment *) XtMalloc(sizeof(XSegment) *
                                    (new_kg->knobg.num_segments + 1));

    new_kg->knobg.move_clockwise = True;
    new_kg->knobg.turning = False;
    new_kg->knobg.timer_id = NULL;
    new_kg->knobg.angle_offset = 0;

   /* Create initial segments */

    (*kc->knobg_class.create_segments)(new_w);

   /* Set up graphic context for use in drawing the knob */

    valuemask = (GCForeground | GCBackground);
    values.foreground = mw->manager.foreground;
    values.background = new_kg->knobg.knob_color;
    new_kg->knobg.gc = XtGetGC((Widget)mw, valuemask, &values); 

   /*  Initialize the interesting input types.  */

   new_kg->gadget.event_mask = XmARM_EVENT | XmACTIVATE_EVENT |
                               XmFOCUS_IN_EVENT | XmFOCUS_OUT_EVENT |
                               XmENTER_EVENT | XmLEAVE_EVENT | XmKEY_EVENT;
}



/************************************************************************
 *
 *  Destroy() -
 *      Clean up allocated resources when the widget is destroyed.
 *        Free any allocated data.
 *
 ************************************************************************/
static void Destroy( Widget w )
{
    XsmKnobGadget kg = (XsmKnobGadget) w;
   
   /* Free graphics contexts */
    XtReleaseGC( w, kg->knobg.gc);
    
   /* Free allocated data */
    XtFree((char *) kg->knobg.segments);

   /* Remove any outstanding timeouts */
    if (kg->knobg.timer_id) XtRemoveTimeOut(kg->knobg.timer_id);

   /* Remove all callbacks */
    XtRemoveAllCallbacks (w, XsmNvalueChangedCallback);
}


/************************************************************************
 *
 *  Resize() -
 *      Redraw according to when size changes.
 *
 ************************************************************************/
static void Resize( Widget w )
{
    XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);

   /* Recreate any segments based on the new size */
    (*kc->knobg_class.create_segments)(w);
}



/************************************************************************
 *
 *  Redisplay() -
 *      Redraw region that go exposed.  Simple and less efficient
 *  expose routines simply redraw the whole widget.  This is an
 *  example of a simple expose routine.
 *
 ************************************************************************/
/* ARGSUSED */
static void Redisplay( Widget w, XEvent *event, Region region)
{
   XsmKnobGadget kg = (XsmKnobGadget) w;
   XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);

  /* Use the class pointer to redraw the whole knob */
   (*kc->knobg_class.draw)(w);

  /* Use the class pointers to highlight or unhighlihht the knob */
   if (kg->gadget.highlighted)
      (*kc->gadget_class.border_highlight)(w);
   else
      (*kc->gadget_class.border_unhighlight)(w);
}



/************************************************************************
 *
 *  SetValues() -
 *      The procedure is used to verify changes to the resource values.
 *        If the resource changes effect the visuals of the widget, the
 *        return value of the procedure should be set to True indicating
 *        that the expose procedure will be called to redraw the widget.
 *        Any changes to the widget's geometry in this procedure will
 *        generate a geometry request to it's parent.
 *
 ************************************************************************/
/* ARGSUSED */
static Boolean SetValues( Widget old_w, Widget request, Widget new_w,
                          ArgList args, Cardinal *num_args )
{
    XsmKnobGadget old_kg = (XsmKnobGadget) old_w;
    XsmKnobGadget new_kg = (XsmKnobGadget) new_w;
    XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(new_w);
    XmManagerWidget mw = (XmManagerWidget) XtParent(new_w);
    Dimension min_width, min_height;
    Boolean redisplay = False; /* flag to determine return value */

   /* verify changes in resource settings */

   /* verify that the maximum value is not less than or equal to zero */
    if (new_kg->knobg.max_val <= 0) {
       XtWarning("MaxValue is must be greater than 0, MaxValue set to 1");
       new_kg->knobg.max_val = 1;
    }

   /*
    * verify that the minimum value is not greater or equal to 
    * the maximum value
    */
    if (new_kg->knobg.min_val >= new_kg->knobg.max_val) {
       XtWarning("MinValue is greater than or equal to MaxValue, MinValue set to 0");
       new_kg->knobg.min_val = 0;
    }

   /* verify that the value is not greater than the maximum value */
    if ((new_kg->knobg.max_val != old_kg->knobg.max_val ||
         new_kg->knobg.value != old_kg->knobg.value) &&
         new_kg->knobg.value > new_kg->knobg.max_val) {
       XtWarning("Value is greater than MaxValue, Changing Value to MaxValue");
       new_kg->knobg.value = new_kg->knobg.max_val;
       redisplay = True;
    }

   /* verify that the value is not less than the minimum value */
    if ((new_kg->knobg.min_val != old_kg->knobg.min_val ||
         new_kg->knobg.value != old_kg->knobg.value) &&
         new_kg->knobg.value < new_kg->knobg.min_val) {
       XtWarning("Value is less than MinValue, Changing Value to MinValue");
       new_kg->knobg.value = new_kg->knobg.min_val;
       redisplay = True;
    }

   /* verify that turn rate is not less than 1 */
    if (new_kg->knobg.turn_delay < 1) {
       XtWarning("Turn Delay is less than 1, Defaulting to 1");
       new_kg->knobg.turn_delay = 1;
    }

   /*
    * Knob margin is a percentage of the total knob,
    * thus it must be between 0 and 100.  Since this resource is
    * typed as a dimension (unsigned short), a negative value
    * becomes a large positive, thus this also checks for < 0/
    */
    if (new_kg->knobg.knob_margin > 100)
    {
       XtWarning("Knob Margin must be between 0 and 100");
       new_kg->knobg.knob_margin = old_kg->knobg.knob_margin;
    }

        
   /* Don't allow a zero size to set */
    if (new_kg->rectangle.width == 0)
       new_kg->rectangle.width = old_kg->rectangle.width;
    if (new_kg->rectangle.height == 0)
       new_kg->rectangle.height = old_kg->rectangle.height;

   /* If the color resources change, the graphic context needs to be updated */
    if (new_kg->knobg.knob_color != old_kg->knobg.knob_color) {
       XGCValues values;
       unsigned long valuemask;

       valuemask = (GCBackground );
       values.background = new_kg->knobg.knob_color;
       XtReleaseGC((Widget)mw, new_kg->knobg.gc);
       new_kg->knobg.gc = XtGetGC((Widget)mw, valuemask, &values); 
       redisplay = True;
    }

    min_width = MIN_KNOB_DIAMETER +
                (2 * (new_kg->knobg.margin_width +
                      new_kg->gadget.highlight_thickness));
    min_height = MIN_KNOB_DIAMETER +
                 (2 * (new_kg->knobg.margin_width +
                       new_kg->gadget.highlight_thickness));
   /* Try to retain a minimum height and width */

    if (new_kg->rectangle.width < min_width)
    {
       new_kg->rectangle.width = min_width;
       redisplay = True;
    }

    if (new_kg->rectangle.height < min_height)
    {
       new_kg->rectangle.height = min_height;
       redisplay = True;
    }

   /*
    * If any of the following resource values changed, the knob
    * segments need to be recomputed and the knob need to be redisplayed
    */
    if (new_kg->knobg.margin_height != old_kg->knobg.margin_height ||
        new_kg->knobg.margin_width != old_kg->knobg.margin_width ||
        new_kg->knobg.knob_margin != old_kg->knobg.knob_margin ||
        new_kg->knobg.max_val != old_kg->knobg.max_val ||
        new_kg->knobg.min_val != old_kg->knobg.min_val ||
        new_kg->rectangle.height != old_kg->rectangle.height ||
        new_kg->rectangle.width != old_kg->rectangle.width ||
        new_kg->gadget.highlight_thickness != 
        old_kg->gadget.highlight_thickness) {

       (*kc->knobg_class.create_segments)(new_w);
       redisplay = True;
    }

   /*  Initialize the interesting input types.  */

    new_kg->gadget.event_mask = XmARM_EVENT | XmACTIVATE_EVENT |
                                XmFOCUS_IN_EVENT | XmFOCUS_OUT_EVENT |
                                XmENTER_EVENT | XmLEAVE_EVENT | XmKEY_EVENT;
    return(redisplay);
}



/************************************************************************
 *
 *  SetValuesAlmost() -
 *      The procedure is used to negotiate a geometry with the parent
 *        if the change in geometry in the SetValues Procedure failed.
 *        Setting the reply->request_mode to zero ends the negotiations.
 *
 ************************************************************************/
/* ARGSUSED */
static void SetValuesAlmost( Widget old_w, Widget new_w,
                             XtWidgetGeometry *request, XtWidgetGeometry *reply)
{
    XsmKnobGadget new_kg = (XsmKnobGadget) new_w;
    XsmKnobGadget old_kg = (XsmKnobGadget) old_w;
    Dimension min_width, min_height;

   /* If the request failed completely, reset old values and return */
    if (reply->request_mode == 0)
    {
       new_kg->knobg.margin_width = old_kg->knobg.margin_width;
       new_kg->knobg.margin_height = old_kg->knobg.margin_height;
       request->request_mode = 0;
       return;
    }

   /* all attempts to accomodate size failed */ 
    if (new_kg->knobg.margin_width == 0 ||
        new_kg->knobg.margin_height == 0)
    {   
       request->request_mode = 0;
       return;
    }

    min_width = MIN_KNOB_DIAMETER +
                (2 * (new_kg->knobg.margin_width +
                      new_kg->gadget.highlight_thickness));
    min_height = MIN_KNOB_DIAMETER +
                 (2 * (new_kg->knobg.margin_width +
                       new_kg->gadget.highlight_thickness));
    
   /* reduce margin to accomodate smaller size */
    if ( reply->width < min_width )
    {
       while ( reply->width < min_width )
       {
          new_kg->knobg.margin_width--;
          min_width -= 2; /* minus one pixel on each side of knob */
          if (new_kg->knobg.margin_width == 0) break;
       }
       request->width = min_width;
    }
    else /* accept larger size */
    {
       request->width = reply->width;
    }

   /* reduce margin to accomodate smaller size */
    if (reply->height < min_height)
    {
       while (reply->height < min_height)
       {
          new_kg->knobg.margin_height--;
          min_height -= 2; /* minus one pixel on each side of knob */
          if (new_kg->knobg.margin_height == 0) break;
       }
       request->height = min_height;
    }
    else /* accept larger size */
    {
       request->width = reply->width;
    }
}

/************************************************************************
 *
 *  QueryGeometry() -
 *      The procedure is used to negotiate a geometry with the parent
 *        when the parent is requesting a preferred of the child.
 *        If the values in the request don't match the values in the
 *        reply, we need to return either XtGeometryNo or XtGeoemtryAlmost.
 *        Use XtGeoemtryNo if the values in the request match the current values.
 *        Note: This procedure is only interested in height and width values.
 *        It will accept any changes to x, y, border width and stacking order.
 *
 ************************************************************************/

static XtGeometryResult QueryGeometry( Widget w, XtWidgetGeometry *request, 
                                            XtWidgetGeometry *reply)
{
    XsmKnobGadget kg = (XsmKnobGadget) w;

    reply->width = kg->knobg.orig_width;
    reply->height = kg->knobg.orig_height;
    reply->request_mode = CWWidth | CWHeight;

   /* return XtGeometryYes if the request matches the reply width and height */
    if (((request->request_mode & CWWidth) && request->width == reply->width) &&
       ((request->request_mode & CWHeight) && request->height == reply->height))
       return XtGeometryYes;

   /* return XtGeometryNo if reply matches our current width and height */
    if (reply->width == kg->rectangle.width && reply->height == kg->rectangle.height)
       return XtGeometryNo; 
    
   /* return XtGeometryAlmost if one of the reply fields doesn't match the
    * current or request width or height.
    */
    return XtGeometryAlmost; 
}



/************************************************************************
 *
 * HighlightKnob
 *        The is a gadget class procedure for drawing a highlight around
 * the knobg.  Knob needs a customized one to draw a circular highlight.
 *
 ************************************************************************/
static void HighlightKnob( Widget w )
{
   XsmKnobGadget kg = (XsmKnobGadget) w;
   XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
   XmManagerWidget mw = (XmManagerWidget) XtParent(w);
   unsigned int outer_diameter, inner_diameter, diameter;
   XGCValues values;
   unsigned long valuemask;
    
  /* Get the diameter */
   (*kc->knobg_class.get_diameters)(w, &outer_diameter, &inner_diameter);

   diameter = outer_diameter + (2 * kg->gadget.highlight_thickness);

   /* Change the graphic context to highlight the knob */
   valuemask = (GCLineWidth | GCForeground);
   values.foreground = mw->manager.highlight_color;
   values.line_width = kg->gadget.highlight_thickness;
   XChangeGC(XtDisplay(kg), kg->knobg.gc, valuemask, &values);

  /* Draw the highlight */
   XDrawArc(XtDisplay(kg), XtWindow(kg), mw->manager.highlight_GC,
            kg->rectangle.x + kg->knobg.margin_width,
            kg->rectangle.y + kg->knobg.margin_height, diameter,
            diameter, 0, 64*360);

  /* Set flags to indicate that the highlight has been drawn */
   kg->gadget.highlighted = True;
   kg->gadget.highlight_drawn = True;

  /* Change the graphic context back for drawing the knob */
   valuemask = (GCLineWidth | GCForeground);
   values.foreground = kg->knobg.knob_color;
   values.line_width = 0;
   XChangeGC(XtDisplay(kg), kg->knobg.gc, valuemask, &values);
}


/************************************************************************
 *
 * UnhighlightKnob
 *        The is a gadget class procedure for erasing a highlight around
 * the knobg.  Knob needs a customized one to erase a circular highlight.
 *
 ************************************************************************/
static void UnhighlightKnob( Widget w )
{
   XsmKnobGadget kg = (XsmKnobGadget) w;
   XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
   unsigned int outer_diameter, inner_diameter, diameter;
   XmManagerWidget mw = (XmManagerWidget) XtParent(w);
   XGCValues values;
   unsigned long valuemask;
    
  /* Get the diameter */
   (*kc->knobg_class.get_diameters)(w, &outer_diameter, &inner_diameter);

   diameter = outer_diameter + (2 * kg->gadget.highlight_thickness);

  /* Change the graphic context to unhighlight */
   valuemask = (GCLineWidth | GCForeground);
   values.foreground = mw->core.background_pixel;
   values.line_width = kg->gadget.highlight_thickness;
   XChangeGC(XtDisplay(kg), kg->knobg.gc, valuemask, &values);

  /* Erase the highlight */
   XDrawArc(XtDisplay(kg), XtWindow(kg), kg->knobg.gc,
            kg->rectangle.x + kg->knobg.margin_width,
            kg->rectangle.y + kg->knobg.margin_height,
            diameter, diameter, 0, 64*360);

   kg->gadget.highlighted = False;
   kg->gadget.highlight_drawn = False;

  /* Change the graphic context back for drawing the knob */
   values.foreground = kg->knobg.knob_color;
   values.line_width = 0;
   XChangeGC(XtDisplay(kg), kg->knobg.gc, valuemask, &values);

}


/************************************************************************
 *
 *     ArmAndActivate
 *	  Upon receipt of an activate key event (spacebar, select key,
 *        return (enter) key, this class function will be called and
 *        it will toggle the direction of the knob turning.
 *
 ************************************************************************/
/* ARGSUSED */
static void
ArmAndActivate( Widget wid, XEvent *event, String *params,
                Cardinal *num_params )
{
       ToggleKnob(wid, event, NULL, NULL);
}



/************************************************************************
 *
 *  InputDispatch
 *     This method catches input sent by a manager and dispatches it
 *     to the individual routines.
 *
 ************************************************************************/
static void
InputDispatch( Widget wid, XEvent *event, Mask event_mask )
{
    XsmKnobGadget kg = (XsmKnobGadget) wid;

    if (event_mask & XmARM_EVENT)
       StartKnobTurn(wid, event, NULL, NULL);
    else if (event_mask & XmACTIVATE_EVENT)
       ReleaseKnob(wid, event, NULL, NULL);
    else if (event_mask & XmKEY_EVENT)
       HandleKey(wid, event, NULL, NULL);
    else if (event_mask & XmENTER_EVENT)
        _XmEnterGadget ((Widget) kg, event, NULL, NULL);
    else if (event_mask & XmLEAVE_EVENT)
        _XmLeaveGadget ((Widget) kg, event, NULL, NULL);
    else if (event_mask & XmFOCUS_IN_EVENT)
        _XmFocusInGadget( (Widget) kg, event, NULL, NULL);
    else if (event_mask & XmFOCUS_OUT_EVENT)
        _XmFocusOutGadget( (Widget) kg, event, NULL, NULL);
}

/************************************************************************
 *
 *  VisualChange
 *      This method is called from XmManagerClass set values when
 *      the managers visuals have changed.  The gadget regenerates any
 *      GC based on the visual changes and returns True indicating a
 *      redraw is needed.  Otherwise, False is returned.
 *
 ************************************************************************/
static Boolean
VisualChange( Widget wid, Widget cmw, Widget nmw )
{
    XsmKnobGadget kg = (XsmKnobGadget) wid ;
    XmManagerWidget curmw = (XmManagerWidget) cmw ;
    XmManagerWidget newmw = (XmManagerWidget) nmw ;
    
    if (curmw->manager.foreground != newmw->manager.foreground)
    {
       XSetForeground(XtDisplay(kg), kg->knobg.gc, newmw->manager.foreground);
       return (True);
    }

    if (curmw->manager.highlight_color != newmw->manager.highlight_color ||
        curmw->core.background_pixel != newmw->core.background_pixel ||
        curmw->manager.highlight_pixmap != newmw->manager.highlight_pixmap)
       return (True);

    return (False);
}


/************************************************************************
 *
 * GetDiameters
 *         This knob class procedure is used to determine the inner and out
 * diameter of the knobg.  The knob draws 3 circles. The outer most circle
 * is for highlighting. The next circle is for drawing the outer edge of
 * the knobg.  The margin_width or the margin_height determine the outer edge
 * of the circle.  The knob_margin determines the inner most circle (being
 * a percentage of the knob itself).
 *
 ************************************************************************/
static void GetDiameters( Widget w, unsigned int *diameter,
                         unsigned int *inner_diameter)
{
    XsmKnobGadget kg = (XsmKnobGadget) w;
    
    if ((Dimension)(kg->rectangle.height - (2 * kg->knobg.margin_width)) >=
        (Dimension)(kg->rectangle.width - (2 * kg->knobg.margin_height)))
       *diameter = kg->rectangle.width - (2 * (kg->knobg.margin_width + 
                                          kg->gadget.highlight_thickness));
    else
       *diameter = kg->rectangle.height - (2 * (kg->knobg.margin_height +
                                           kg->gadget.highlight_thickness));

    if (*diameter < MIN_KNOB_DIAMETER) *diameter = MIN_KNOB_DIAMETER;

    *inner_diameter = (*diameter * kg->knobg.knob_margin)/100;

    if (*inner_diameter < 1) *inner_diameter = 1;
}



/************************************************************************
 *
 * CreateSegments
 *         This knob class procedure is used to create the line segments used
 * in drawing the knobg.
 *
 ************************************************************************/
static void CreateSegments( Widget w )
{
    XsmKnobGadget kg = (XsmKnobGadget) w;
    XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
    unsigned int diameter, inner_diameter;
    unsigned int radius, inner_radius;
    double    angle, cosine, sine, angle_change;
    Position center_x, center_y;
    int       i;
    XSegment   *ptr;
    
    ptr = kg->knobg.segments;
    
    angle_change = RADIANS(MAX_ANGLE) /(float)(kg->knobg.num_segments - 1);

    (*kc->knobg_class.get_diameters)(w, &diameter, &inner_diameter);

    angle = kg->knobg.angle_offset;

    center_x = kg->rectangle.x + (diameter >> 1) + kg->knobg.margin_width +
               kg->gadget.highlight_thickness;
    center_y = kg->rectangle.y + (diameter >> 1) + kg->knobg.margin_height +
               kg->gadget.highlight_thickness; 

    radius = ((diameter >> 1) * 80)/100;

    if (inner_diameter > 2)
       inner_radius = ((inner_diameter >> 1) * 110)/100;
    else
       inner_radius = 1;

    for (i = 0; i < kg->knobg.num_segments; i++) {
       cosine = cos(angle);
       sine   = sin(angle);
       ptr->x1   = (short)(center_x + radius * sine);
       ptr->y1 = (short)(center_y - radius * cosine);
       ptr->x2   = (short)(center_x + inner_radius * sine);
       ptr++->y2 = (short)(center_y - inner_radius * cosine);
       angle += angle_change;
    }
}



/************************************************************************
 *
 * DrawIndicator
 *         This knob class procedure is used to draw the indicator portion of
 * the knobg.
 *
 ************************************************************************/
static void DrawIndicator( Widget w )
{
   XsmKnobGadget kg = (XsmKnobGadget) w;
   XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
   unsigned int diameter, inner_diameter;
   unsigned int radius;
   double angle;
   Position center_x, center_y;

   (*kc->knobg_class.get_diameters)(w, &diameter, &inner_diameter);

   angle = (RADIANS(MAX_ANGLE) * kg->knobg.value)/ 
                (kg->knobg.max_val - kg->knobg.min_val) ;

   center_x = kg->rectangle.x + (diameter >> 1) + kg->knobg.margin_width +
               kg->gadget.highlight_thickness;
   center_y = kg->rectangle.y + (diameter >> 1) + kg->knobg.margin_height +
               kg->gadget.highlight_thickness;

   radius = ((diameter >> 1) * 93)/100;

   diameter = (diameter * 5)/100;
   if (diameter == 0) diameter = 1;

   center_x += (Position)(radius * sin(angle)) - (diameter >> 1);
   center_y -= (Position)(radius * cos(angle)) + (diameter >> 1);


   XDrawArc(XtDisplay(w), XtWindow(w), kg->knobg.gc, 
            center_x, center_y, diameter, diameter, 0, 64*360);

   XFillArc(XtDisplay(w), XtWindow(w), kg->knobg.gc,
            center_x, center_y, diameter, diameter, 0, 64*360);

}


/************************************************************************
 *
 * TurnKnob
 *         This knob class procedure is used in the timeout procedures
 * to show the knob turning.
 *
 ************************************************************************/
/* ARGSUSED */
static void TurnKnob( XtPointer closure, XtIntervalId *id )
{
   Widget w = (Widget) closure;
   XsmKnobGadget kg = (XsmKnobGadget) w;
   XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
   XmManagerWidget mw = (XmManagerWidget) XtParent(w);
   XsmKnobGadgetCallbackStruct cb;

   if ((kg->knobg.value < kg->knobg.max_val - 1 && kg->knobg.move_clockwise) ||
       (kg->knobg.value > 1 && !kg->knobg.move_clockwise)) {
       kg->knobg.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                           (unsigned long) kg->knobg.turn_delay,
                                           kc->knobg_class.turn, (XtPointer) kg);
   } else {
       kg->knobg.turning = False;
   }

   XSetForeground(XtDisplay(kg), kg->knobg.gc, kg->knobg.knob_color);
   (*kc->knobg_class.draw_indicator)(w);

   XDrawSegments(XtDisplay(w), XtWindow(w), kg->knobg.gc,
                 kg->knobg.segments, kg->knobg.num_segments);

   if (kg->knobg.move_clockwise)
      kg->knobg.value++;
   else
      kg->knobg.value--;

   if (kg->knobg.value_changed_callback) {
      cb.reason = XsmCR_VALUE_CHANGED;
      cb.event = NULL;
      cb.value = kg->knobg.value;
      XtCallCallbackList(w, kg->knobg.value_changed_callback, (XtPointer) &cb);
   }

   kg->knobg.angle_offset = (RADIANS(MAX_ANGLE) * kg->knobg.value)/ 
                            (kg->knobg.max_val - kg->knobg.min_val) ;
   (*kc->knobg_class.create_segments)(w);

   XSetForeground(XtDisplay(kg), kg->knobg.gc, kg->knobg.indicator_color);
   (*kc->knobg_class.draw_indicator)(w);

   XSetForeground(XtDisplay(kg), kg->knobg.gc, mw->manager.foreground);
   XDrawSegments(XtDisplay(w), XtWindow(w), kg->knobg.gc,
                 kg->knobg.segments, kg->knobg.num_segments);

}


/************************************************************************
 *
 * DrawKnob
 *         This knob class procedure is used to draw the knobg.
 *
 ************************************************************************/
static void DrawKnob( Widget w )
{
   XsmKnobGadget kg = (XsmKnobGadget) w;
   XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
   XmManagerWidget mw = (XmManagerWidget) XtParent(w);
   unsigned int diameter, inner_diameter, top_knob_start;

   (*kc->knobg_class.get_diameters)(w, &diameter, &inner_diameter);
    
   /* bottom of knob */ 
   XSetForeground(XtDisplay(kg), kg->knobg.gc, kg->knobg.knob_color);
   XFillArc(XtDisplay(kg), XtWindow(kg), kg->knobg.gc,
            kg->rectangle.x +
            (kg->knobg.margin_width + kg->gadget.highlight_thickness), 
            kg->rectangle.y +
            (kg->knobg.margin_height + kg->gadget.highlight_thickness), 
            diameter, diameter, 0, 64*360);
   XSetForeground(XtDisplay(kg), kg->knobg.gc, mw->manager.foreground);

   XDrawArc(XtDisplay(kg), XtWindow(kg), kg->knobg.gc, 
            kg->rectangle.x +
            (kg->knobg.margin_width + kg->gadget.highlight_thickness), 
            kg->rectangle.y +
            (kg->knobg.margin_height + kg->gadget.highlight_thickness), 
            diameter, diameter, 0, 64*360);

   top_knob_start = (diameter - ((diameter * kg->knobg.knob_margin)/100))/2;

   /* top of knob */ 
   XDrawArc(XtDisplay(kg), XtWindow(kg), kg->knobg.gc,
            kg->rectangle.x + (top_knob_start +
            kg->knobg.margin_width + kg->gadget.highlight_thickness), 
            kg->rectangle.y + (top_knob_start +
            kg->knobg.margin_height + kg->gadget.highlight_thickness), 
            inner_diameter, inner_diameter, 0, 64*360);

   XDrawSegments(XtDisplay(w), XtWindow(w), kg->knobg.gc,
                 kg->knobg.segments, kg->knobg.num_segments);

   XSetForeground(XtDisplay(kg), kg->knobg.gc, kg->knobg.indicator_color);
   (*kc->knobg_class.draw_indicator)(w);
   XSetForeground(XtDisplay(kg), kg->knobg.gc, mw->manager.foreground);

}

/*********************************************************************
 *
 * Callproc procedures defined to set default settings for resources
 *
 *********************************************************************/

/******************************************************************
 *  DefaultKnobColor
 *       Get the default knob color from parent's background
 *       if the parent is a manager widget otherwise get the
 *       Motif default background color.
 ******************************************************************/
static void
DefaultKnobColor( Widget g, int offset, XrmValue *value )
{
        XmManagerWidget         mw = (XmManagerWidget) XtParent (g);
        static Pixel            pixel;

        value->addr = (XtPointer) &pixel;
        value->size = sizeof (Pixel);

        if (XmIsManager((Widget) mw))
           pixel = mw->core.background_pixel;
        else
        {
          /* Work around for bug in _XmBackgroundColorDefault callproc */
           _XmSetDefaultBackgroundColorSpec(XtScreen(g), XmDEFAULT_BACKGROUND);

           _XmBackgroundColorDefault (g, offset, value);
        }
}


/******************************************************************
 *  DefaultIndicatorColor
 *       Get the default indicator color from parent's foreground
 *       if the parent is a manager widget otherwise get the
 *       Motif default background color.
 ******************************************************************/
static void
DefaultIndicatorColor( Widget g, int offset, XrmValue *value )
{
        XmManagerWidget         mw = (XmManagerWidget) XtParent (g);
        static Pixel            pixel;

        value->addr = (XtPointer) &pixel;
        value->size = sizeof (Pixel);

        if (XmIsManager ((Widget) mw))
           pixel = mw->manager.foreground;
        else
           _XmForegroundColorDefault (g, offset, value);
}


/****************************************************************
 *
 * Functions defined in the action table.
 *
 ****************************************************************/


/************************************************************************
 *
 * HandleKey
 *        An action procedure that is called on any key press.
 *        This procedure will toggle the key turning in the direction
 *        of the key pressed, '<' turns left and '>' turns right.
 *
 ************************************************************************/
/* ARGSUSED */
static void HandleKey( Widget w, XEvent *event, char **params,
                      Cardinal *num_params )
{
    KeySym keysym;
    
   /* Get the keysym from the event */
    XLookupString((XKeyEvent *) event, (char *)NULL, 0, &keysym, NULL);

    if (keysym == XK_less)
       ToggleLeft( w, event, params, num_params );
    else if (keysym == XK_greater)
       ToggleRight( w, event, params, num_params );
}



/* ARGSUSED */
/************************************************************************
 *
 * ToggleLeft
 *        An action procedure that is called on a key press.
 *        This procedure will add a timeout to start the knob turning left.
 *
 ************************************************************************/
static void ToggleLeft( Widget w, XEvent *event, char **params,
                        Cardinal *num_params )
{
    XsmKnobGadget kg = (XsmKnobGadget) w;
    XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
    
   /* If the knob has not started turning, added the timeout */
    if (!kg->knobg.turning) {
       if (kg->knobg.value > kg->knobg.min_val) {
          kg->knobg.move_clockwise = False;
          kg->knobg.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kg->knobg.turn_delay,
                                          kc->knobg_class.turn, (XtPointer) w);
          kg->knobg.turning = True;
       }
    } else {
      /*
       * If it is already turning, remove the old timeout and
       * continue turning with a new timeout
       */
       if (kg->knobg.timer_id) XtRemoveTimeOut(kg->knobg.timer_id);

      /* if it was turning left continue the turning, otherwise stop. */
       if (kg->knobg.move_clockwise) {
          kg->knobg.move_clockwise = False;
          kg->knobg.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kg->knobg.turn_delay,
                                          kc->knobg_class.turn, (XtPointer) w);
       } else {
          kg->knobg.turning = False;
       }
    }
}


/************************************************************************
 *
 * ToggleRight
 *        An action procedure that is called on a key press.
 *        This procedure will add a timeout to start the knob turning right.
 *
 ************************************************************************/
/* ARGSUSED */
static void ToggleRight( Widget w, XEvent *event, char **params,
                         Cardinal *num_params )
{
    XsmKnobGadget kg = (XsmKnobGadget) w;
    XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
    
   /* If the knob has not started turning, added the timeout */
    if (!kg->knobg.turning) {
       if (kg->knobg.value < kg->knobg.max_val) {
          kg->knobg.move_clockwise = True;
          kg->knobg.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kg->knobg.turn_delay,
                                          kc->knobg_class.turn, (XtPointer) w);
          kg->knobg.turning = True;
       }
    } else {
      /*
       * If it is already turning, remove the old timeout and
       * continue turning with a new timeout
       */
       if (kg->knobg.timer_id) XtRemoveTimeOut(kg->knobg.timer_id);

      /* if it was turning right continue the turning, otherwise stop. */
       if (!kg->knobg.move_clockwise) {
          kg->knobg.move_clockwise = True;
          kg->knobg.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kg->knobg.turn_delay,
                                          kc->knobg_class.turn, (XtPointer) w);
       } else {
          kg->knobg.turning = False;
       }
    }
}



/************************************************************************
 *
 * ToggleKnob
 *        An action procedure that is called on a spacebar, return (enter)
 *        or select key press.  This procedure will toggle the direction
 *        for turning.
 *
 ************************************************************************/
/* ARGSUSED */
static void ToggleKnob( Widget w, XEvent *event, char **params,
                      Cardinal *num_params )
{
    XsmKnobGadget kg = (XsmKnobGadget) w;
    
    if (kg->knobg.move_clockwise)
       kg->knobg.move_clockwise = False;
    else
       kg->knobg.move_clockwise = True;
}



/************************************************************************
 *
 * StartKnobTurn
 *        An action procedure that is called on a mouse button press.
 *        This procedure will use a timeout to start the knob from turning.
 *
 ************************************************************************/
/* ARGSUSED */
static void StartKnobTurn( Widget w, XEvent *event, char **params,
                      Cardinal *num_params )
{
    XsmKnobGadget kg = (XsmKnobGadget) w;
    XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
    
   /*
    * As long as the value has not reached the minimum value,
    * continue turning
    */
    if (kg->knobg.move_clockwise)
    {
       if (kg->knobg.value < kg->knobg.max_val)
          kg->knobg.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kg->knobg.turn_delay,
                                          kc->knobg_class.turn, (XtPointer) w);
    }
    else
    {
       if (kg->knobg.value > kg->knobg.min_val)
          kg->knobg.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kg->knobg.turn_delay,
                                          kc->knobg_class.turn, (XtPointer) w);
    }
   
}



/************************************************************************
 *
 * ReleaseKnob
 *        An action procedure that is called on a mouse button release.
 *        This procedure will remove a timeout to stop the knob from turning.
 *
 ************************************************************************/
/* ARGSUSED */
static void ReleaseKnob( Widget w, XEvent *event, char **params,
                         Cardinal *num_params )
{
    XsmKnobGadget kg = (XsmKnobGadget) w;

   /* remove the timeout */
    if (kg->knobg.timer_id) XtRemoveTimeOut(kg->knobg.timer_id);

    kg->knobg.turning = False;
}


/****************************************************************
 *
 * Public Functions
 *
 ****************************************************************/

/********************************************************
 *
 * XsmCreatKnobGadget() -
 *        Knob widget creation convienence routine.
 *
 ********************************************************/

Widget XsmCreateKnobGadget( Widget parent, char *name, ArgList arglist,
                      Cardinal argcount )
{
    return (XtCreateWidget(name, xsmKnobGadgetClass,
                           parent, arglist, argcount));
}

/**********************************************************************
 *
 * XsmGetKnobGadgetValue() -
 *        A function to quickly get the knob value.  If a resource value
 *        needs to be retrieved frequently, it might be a good idea to
 *        have a convenience function like this to avoid costly calls to
 *        XtGetValues().  Too many of these functions can be confusing
 *        the application developer.
 *
 ***********************************************************************/
int XsmGetKnobGadgetValue( Widget w )
{
    XsmKnobGadget kg = (XsmKnobGadget) w;

    return(kg->knobg.value);
}


/***********************************************************************
 *
 * XsmSetKnobGValue() -
 *        A function to quickly set the knob value.  If a resource value
 *        needs to be changed frequently, it might be a good idea to
 *        have a convenience function like this to avoid costly calls to
 *        XtSetValues().  Too many of these functions can be confusing
 *        the application developer.
 *
 ***********************************************************************/
void XsmSetKnobGadgetValue( Widget w, int value )
{
   XsmKnobGadget kg = (XsmKnobGadget) w;
   XsmKnobGadgetClass kc = (XsmKnobGadgetClass) XtClass(w);
   XmManagerWidget mw = (XmManagerWidget) XtParent(w);

  /* return if the value doesn't change */

   if (value == kg->knobg.value) return;

  /* verify the changed values */

   if (value > kg->knobg.max_val) {
      XtWarning("Value is less than MaxValue, Defaulting to MaxValue");
      kg->knobg.value = kg->knobg.max_val;
   }

   if (value < kg->knobg.min_val) {
      XtWarning("Value is greater than MinValue, Defaulting to MinValue");
      kg->knobg.value = kg->knobg.min_val;
   }

  /*
   * If the knob is realized, redraw the knob segments and indicator to
   * indicate the new value
   */

   if (XtIsRealized(w)) {
     /* erase old indicator */
      XSetForeground(XtDisplay(kg), kg->knobg.gc, kg->knobg.knob_color);
      (*kc->knobg_class.draw_indicator)(w);
      XDrawSegments(XtDisplay(w), XtWindow(w), kg->knobg.gc,
                    kg->knobg.segments, kg->knobg.num_segments);

     /* set new value */
      kg->knobg.value = value;
   
     /* recreate segments */
      kg->knobg.angle_offset = (RADIANS(MAX_ANGLE) * kg->knobg.value)/ 
                               (kg->knobg.max_val - kg->knobg.min_val) ;
      (*kc->knobg_class.create_segments)(w);

     /* redraw indicator */
      XSetForeground(XtDisplay(kg), kg->knobg.gc, kg->knobg.indicator_color);
      (*kc->knobg_class.draw_indicator)(w);

     /* redraw segments */
      XSetForeground(XtDisplay(kg), kg->knobg.gc, mw->manager.foreground);
      XDrawSegments(XtDisplay(w), XtWindow(w), kg->knobg.gc,
                    kg->knobg.segments, kg->knobg.num_segments);
   } else {
      kg->knobg.value = value;
   }

}

